# File main.py:
# Defines activation and error functions
# Initializes and trains networks

import numpy as np
import mycnn # See mycnn.py
import time

trial = 6

def newFunc(z):
    return (z+np.sqrt(z**2+1))*0.5
def newFunc2(z):
    return (z+np.sqrt(z**2+0.1))*0.5
def newFunc3(z):
    return (z+np.sqrt(z**2+0.01))*0.5
def newFuncPrime(z, a):
    return a/(2*a-z)

def relu(z):
    return np.maximum(z, 0)
def reluPrime(z, a):
    return (np.sign(z)+1)/2
    
def softplus(z):
    return np.log(1+np.exp(z))
def softplusPrime(z, a):
    return 1/(1+np.exp(-z))

def logistic(z):
    return 1/(1+np.exp(-z))
def logisticPrime(z, a):
    return a*(1-a)

def crossEntropy(y, yHat):
    return -np.sum(y*np.log(yHat))
def crossEntropyPrime(y, yHat):
    return -y/yHat

def arraysFromFiles(XFile, yFile, b):

    XString = XFile.read(b*28*28)
    YString = yFile.read(b)
    if XString == '':
        raise ValueError
    
    xCrop = np.random.randint(5)
    yCrop = np.random.randint(5)

    XRaw = np.fromstring(XString, dtype=np.uint8)
    XShaped = np.transpose(XRaw.astype(float).reshape((b, 1, 28, 28)), (2, 3, 1, 0))
    newX = XShaped[yCrop:yCrop+24,xCrop:xCrop+24]/256

    YLabels = np.fromstring(YString, dtype=np.uint8)
    newY = np.zeros((10, b))
    newY[YLabels, np.arange(b)] = 1
    return newX, newY

def getMNIST(b):

    XTrain = open('../data/train-images-idx3-ubyte', 'r')
    yTrain = open('../data/train-labels-idx1-ubyte', 'r')
    XTest = open('../data/t10k-images-idx3-ubyte', 'r')
    yTest = open('../data/t10k-labels-idx1-ubyte', 'r')

    XTrain.seek(16)
    yTrain.seek(8)
    XTest.seek(16)
    yTest.seek(8)

    t = True
    XFile = XTrain
    yFile = yTrain
        
    while True:
        try:
            newX, newY = arraysFromFiles(XFile, yFile, b)
            yield newX, newY, t
        except ValueError:
            XFile.seek(16)
            yFile.seek(8)
            t = not t
            if t:
                XFile = XTrain
                yFile = yTrain
            else:
                XFile = XTest
                yFile = yTest

#x = [[newFunc, newFuncPrime], [relu, reluPrime], [softplus, softplusPrime], [logistic, logisticPrime]]
x = [[newFunc2, newFuncPrime], [newFunc3, newFuncPrime]]

netList = []

numNets = 2
batch = 20
testBatchesPerEpoch = 10000/batch
numEpochs = 50

accuracyData = np.ndarray((numEpochs, numNets))
times = np.ndarray(numNets)

for a in range(numNets):
    allLayers = [{'type': 'conv', 'kernel':   5, 'neurons': 8},
                 {'type': 'act',  'f': x[a][0], 'fPrime': x[a][1]},
                 {'type': 'pool', 'factor':   2},
                 {'type': 'conv', 'kernel':   5, 'neurons': 16},
                 {'type': 'act',  'f': x[a][0], 'fPrime': x[a][1]},
                 {'type': 'pool', 'factor':   2},
                 {'type': 'fc',   'shape':  (10,)},
                 {'type': 'softmax'}]
    newNet = mycnn.CNN(24, 1, batch, crossEntropy, crossEntropyPrime)
    netList.append(newNet)
    for layer in allLayers:
        newNet.addLayer(layer)

    numIterations = 0
    correctInEpoch = 0
    

    start = time.time()
    for datum in newNet.train(getMNIST):
        correctInEpoch += datum
        numIterations += 1
        if numIterations%testBatchesPerEpoch == 0:
            currentEpoch = numIterations/testBatchesPerEpoch
            accuracyData[currentEpoch-1][a] = correctInEpoch
            correctInEpoch = 0
        if numIterations == testBatchesPerEpoch*numEpochs:
            times[a] = time.time()-start
            break

np.savetxt('../results/accuracy_data_%d.txt'%trial, accuracyData, delimiter=',', fmt='%05u')
np.savetxt('../results/times_%d.txt'%trial, times, fmt='%05u')
